/*
 * bg.h
 * Copyright (C) 2019 Peter Belkner <pbelkner@users.sf.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2.0 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 */
#if ! defined (__BG_H__) // [
#if defined (HAVE_CONFIG_H) // [
#include <config.h>
#endif // ]
#include <stdio.h>
#include <pbutil_priv.h>
#include <ff.h>

///////////////////////////////////////////////////////////////////////////////
#define BG_LIB1770
#if defined (BG_LIB1770) // [
#include <lib1770.h>

///////////////////////////////////////////////////////////////////////////////
#define BG_CHANNEL_LFE 3
#endif // ]

#if defined (__cplusplus) // [
extern "C" {
#endif // ]

///////////////////////////////////////////////////////////////////////////////
//#define BG_LIST
//#define BG_ANNOTATOR
#define BG_OVERWRITE
#define BG_INTERVAL
#define BG_MUXER_PATH_TEMP
#define BG_FILE
#define BG_TREE_TYPE
#define BG_TEMP_PREFIX FFL(".")
#define BG_SKIP_FOLDER
//#define BG_ERROR_LENGTH
// TODO: disable the following:
//#define BG_EXTENSIONS
#if defined (BG_LIB1770) // [
#define BG_PRINT
#endif // ]
#if defined (BG_PRINT) // [
#endif // ]
//#define BG_BWF_TAGS
//#define BG_TRACK_OPTIONS
//#define BG_TREE_CREATE_CHILD_WARNING
#if defined (_WIN32) // [
#define BG_WIN32_TARGET_UTF8
// WARNING:  define BG_WIN32_CREATE_LOCALE at your own risk: at the run-time
//    of our system _create_locale() doesn't exist and hence you're compiling
//    un-tested code.
//#define BG_WIN32_CREATE_LOCALE
#else // ] [
//#define BG_TYPE_DIR
#endif // ]

///////////////////////////////////////////////////////////////////////////////
typedef unsigned long bg_bits_t;
typedef struct bg_album bg_album_t;
typedef enum bg_track_tag bg_track_tag_t;
typedef struct bg_track_target bg_track_target_t;
typedef struct bg_track bg_track_t;
#if defined (BG_TREE_TYPE) // [
typedef enum bg_tree_type bg_tree_type_t;
#endif // ]
typedef const struct bg_tree_vtbl bg_tree_vtbl_t;
typedef struct bg_tree_path bg_tree_path_t;
#if defined (_WIN32) // [
typedef struct bg_tree_patha bg_tree_patha_t;
#endif // ]
typedef struct bg_tree bg_tree_t;

typedef const struct bg_visitor_vtbl bg_visitor_vtbl_t;
typedef struct bg_visitor bg_visitor_t;

#if defined (BG_LIB1770) // [
typedef struct bg_param_block bg_param_block_t;
typedef enum bg_flags_agg bg_flags_agg_t;
#endif // ]
typedef enum bg_flags_ext bg_flags_ext_t;
typedef enum bg_flags_mode bg_flags_mode_t;
#if defined (BG_LIB1770) // [
typedef enum bg_flags_norm bg_flags_norm_t;
#endif // ]
#if defined (BG_PRINT) // [
typedef const struct bg_print_vmt bg_print_vmt_t;
typedef const struct bg_print_conf bg_print_conf_t;
#endif // ]
typedef struct bg_param bg_param_t;

///////////////////////////////////////////////////////////////////////////////
#if defined (BG_INTERVAL) // [
int64_t bg_parse_time(const ffchar_t *s);
#endif // ]
ffchar_t *bg_pathnorm(ffchar_t *path);
const ffchar_t *bg_basename(const ffchar_t *path);
#if defined (_WIN32) // [
char *bg_wcs2str(const wchar_t *wcs, unsigned int codepage);
const char *bg_basenamea(const char *path);
#endif // ]

///////////////////////////////////////////////////////////////////////////////
#if defined (BG_TREE_TYPE) // [
enum bg_tree_type {
  BG_TREE_TYPE_NULL,
#if defined (BG_FILE) // [
  BG_TREE_TYPE_FILE,
#endif // ]
  BG_TREE_TYPE_TRACK,
  BG_TREE_TYPE_ALBUM,
  BG_TREE_TYPE_ROOT/*=BG_TREE_TYPE_ALBUM*/,
};
#endif // ]

struct bg_tree_vtbl {
#if defined (PBU_DEBUG) // [
  const char *id;
#endif // ]
#if defined (BG_TREE_TYPE) // [
  const bg_tree_type_t type;
#endif // ]
  void (*destroy)(bg_tree_t *tree);
  int (*accept)(bg_tree_t *tree, bg_visitor_t *vis);
};

////////
int bg_root_create(bg_tree_t *tree, bg_param_t *param);

#if ! defined (BG_ANNOTATOR) // [
int bg_root_annotation_create(bg_tree_t *tree);
void bg_root_annotation_destroy(bg_tree_t *tree);
#endif // ]

#if defined (_WIN32) // [
int bg_root_append(bg_tree_t *tree, ffchar_t *path);
#else // ] [
int bg_root_append(bg_tree_t *tree, const ffchar_t *path);
#endif // ]

////////
struct bg_album {
  unsigned long count;
  bg_tree_t *first;
};

int bg_album_create(bg_tree_t *tree, bg_param_t *param);
void bg_album_destroy(bg_tree_t *tree);

#if ! defined (BG_ANNOTATOR) // [
int bg_album_annotation_create(bg_tree_t *tree);
void bg_album_annotation_destroy(bg_tree_t *tree);
#endif // ]

const ffchar_t *bg_album_target_purge(bg_tree_t *tree);

////////
enum bg_track_tag {
  BG_TRACK_TAG_ALGORITHM=1<<0,
  BG_TRACK_TAG_REFERENCE_LOUDNESS=1<<1,
  BG_TRACK_TAG_TRACK_GAIN=1<<2,
  BG_TRACK_TAG_TRACK_PEAK=1<<3,
  BG_TRACK_TAG_ALBUM_GAIN=1<<4,
  BG_TRACK_TAG_ALBUM_PEAK=1<<5,
  BG_TRACK_TAG_MAX=1<<6,
};

////////
struct bg_track_target {
#if defined (_WIN32) // [
#if defined (BG_WIN32_TARGET_UTF8) // [
  struct {
#endif // ]
    char *path;
    const char *basename;
#if defined (BG_WIN32_TARGET_UTF8) // [
  } utf8;
#endif // ]

  // if LANG is set to e.g. "en_US.UTF-8" we assume we're run from
  // e.g. MSYS2 shell undestanding UTF-8 otherwise from MS console using
  // codepage OEM. In the latter case we need an OEM representation of
  // e.g. basename.
  struct {
    char *basename;
  } oem;
#endif // ]
#if defined (BG_ERROR_LENGTH) // [
  int preprocessed;
#endif // ]
  wchar_t *title;

  struct {
    size_t len;
  } pfx;
};

////////
struct bg_track {
  ff_inout_t input;
  ff_analyzer_t analyzer;
#if defined (BG_TRACK_OPTIONS) // [
  AVDictionary *options;
#endif // ]
#if defined (FF_INPUT_OPEN_SFX) // [
  ff_param_decode_t decode;
#endif // ]
	bg_track_target_t target;

#if defined (BG_MUXER_PATH_TEMP) // [
  // an utf-8 representation of the target path.
  struct {
    char *path;
  } temp;
#endif // ]


  struct {
    unsigned long id;
  } root;

  struct {
    unsigned long id;
  } album;

#if defined (BG_LIB1770) // [
  struct {
    lib1770_pre_t *pre;
  } filter;

  struct {
    lib1770_block_t *momentary;
    lib1770_block_t *shortterm;
  } block;
#endif // ]
};

int bg_track_create(bg_tree_t *tree);

#if ! defined (BG_ANNOTATOR) // [
#if defined (FF_INPUT_OPEN_SFX) // [
int bg_track_annotation_create(bg_tree_t *tree, const ffchar_t *sfx);
#else // ] [
int bg_track_annotation_create(bg_tree_t *tree);
#endif // ]
void bg_track_annotation_destroy(bg_tree_t *tree);
#endif // ]
#if defined (FF_OUTPUT_CALLBACK_SFX) // [
// from bg_annotator.c:
const ffchar_t *bg_track_out_sfx(const bg_track_t *track,
    const bg_tree_t *tree);
#endif // ]

#if defined (BG_FILE) // [
int bg_file_create(bg_tree_t *tree);

#if ! defined (BG_ANNOTATOR) // [
int bg_file_annotation_create(bg_tree_t *tree);
void bg_file_annotation_destroy(bg_tree_t *tree);
#endif // ]
#endif // ]

////////
struct bg_tree_path {
  ffchar_t *path;
  const ffchar_t *basename;
};

int bg_tree_source_create(bg_tree_path_t *tp, bg_tree_t *parent,
    const ffchar_t *path);
void bg_tree_path_destroy(bg_tree_path_t *tp);

#if defined (_WIN32) // [
////////
struct bg_tree_patha {
  char *path;
  const char *basename;
};

int bg_tree_patha_create(bg_tree_patha_t *p, const wchar_t *path,
    unsigned int codepage);
void bg_tree_patha_destroy(bg_tree_patha_t *p);
#endif // ]

////////
struct bg_tree {
  bg_tree_vtbl_t *vtbl;
  bg_param_t *param;

  bg_tree_path_t source;

#if defined (_WIN32) // [
  // if LANG is set to e.g. "en_US.UTF-8" we assume we're run from
  // e.g. MSYS2 shell undestanding UTF-8 otherwise from MS console using
  // codepage OEM. In the latter case we need an OEM representation of
  // e.g. basename.
  struct {
    char *basename;
  } oem;

	bg_tree_patha_t utf8;
#endif // ]
  bg_tree_path_t target;
#if defined (BG_MUXER_PATH_TEMP) // [
  bg_tree_path_t temp;
#endif // ]
  bg_tree_t *root;
  bg_tree_t *parent;
  bg_tree_t *next;

#if defined (BG_LIB1770) // [
  struct {
    lib1770_stats_t *momentary;
    lib1770_stats_t *shortterm;
    double samplepeak;
    double truepeak;
  } stats;
#endif // ]

  union {
    bg_album_t album;
    bg_track_t track;
  };
};

int bg_tree_common_create(bg_tree_t *tree, bg_param_t *param,
    bg_tree_t *root, bg_tree_t *parent, const ffchar_t *path,
    bg_tree_vtbl_t *vtbl);
void bg_tree_common_destroy(bg_tree_t *tree);

// narrow character representation of the basename.
const char *bg_tree_in_basanamen(bg_tree_t *tree);
const char *bg_tree_out_basanamen(bg_tree_t *tree);
#if defined (_WIN32) // [
// wide character representation of the basename.
const wchar_t *bg_tree_in_basanamew(bg_tree_t *tree);
const wchar_t *bg_tree_out_basanamew(bg_tree_t *tree);
#endif // ]

#if defined (BG_TYPE_DIR) // [
int bg_child_create(bg_tree_t **tree, bg_param_t *param, bg_tree_t *root,
    bg_tree_t *parent, const ffchar_t *path, int dir);
#else // ] [
int bg_child_create(bg_tree_t **tree, bg_param_t *param, bg_tree_t *root,
    bg_tree_t *parent, const ffchar_t *path);
#endif // ]
void bg_child_destroy(bg_tree_t **tree);

#if defined (BG_LIB1770) // [
double *bg_tree_samplepeak(bg_tree_t *tree);
double *bg_tree_truepeak(bg_tree_t *tree);
int bg_tree_merge(bg_tree_t *lhs, const bg_tree_t *rhs);
#if ! defined (BG_PRINT) // [
void bg_tree_result(const bg_tree_t *tree, FILE *f);
#endif // ]
#endif // ]

///////////////////////////////////////////////////////////////////////////////
// several phases are implemented by means of the visitor pattern (cf. e.g.
// https://en.wikipedia.org/wiki/Visitor_pattern.) we've got the following
// phases:
// 1) annotation (cf. below),
// 2) analysis, and
// 3) re-muxing/transcoding.
struct bg_visitor_vtbl {
#if defined (PBU_DEBUG) // [
  const char *id;
#endif // ]
  void (*destroy)(bg_visitor_t *vis);
#if defined (BG_FILE) // [
  int (*dispatch_file)(bg_visitor_t *vis, bg_tree_t *tree);
#endif // ]
  int (*dispatch_track)(bg_visitor_t *vis, bg_tree_t *tree);
  int (*dispatch_album)(bg_visitor_t *vis, bg_tree_t *tree);
  int (*dispatch_root)(bg_visitor_t *vis, bg_tree_t *tree);
};

#if defined (BG_ANNOTATOR) // [
// when re-muxing/transcoding is enabled we run an annotation phase
// with the aim of creating the tree's output path representation, i.e.
// 1) char/wchar_t (ffchar_t) representation (on linux and windows),
// 2) utf-8 represention (just on windows), and
// 3) oem representation (just on windows)
// as early as possible in order to flag possible errors before running
// the time consuming analysis.
// NOTE:  the annotation phase is really needed because it may re-organize
//   the target tree, cf. bg_album_target_purge()!
int bg_annotator_create(bg_visitor_t *vis);
#endif // ]
int bg_analyzer_create(bg_visitor_t *vis);
int bg_muxer_create(bg_visitor_t *vis);

////////
struct bg_visitor {
  bg_visitor_vtbl_t *vtbl;
  int depth;
};

///////////////////////////////////////////////////////////////////////////////
#if defined (BG_PRINT) // [
struct bg_print_vmt {
  const char *id;
  int infix;
  void (*encoding)(bg_param_t *param, int bits);
  int (*head)(bg_tree_t *tree, int depth, FILE *f);
  int (*tail)(bg_tree_t *tree, int depth, FILE *f);
};

extern bg_print_vmt_t bg_print_classic_vmt;
extern bg_print_vmt_t bg_print_xml_vmt;
#endif // ]

#if defined (BG_LIB1770) // [
////////
struct bg_param_block {
  double ms;
  int partition;

  struct {
    double gate;
  } mean;

  struct {
    double gate;
    double lower_bound;
    double upper_bound;
  } range;
};

////////
enum {
  BG_FLAGS_AGG_MOMENTARY_MEAN_OFFSET,
  BG_FLAGS_AGG_MOMENTARY_MAXIMUM_OFFSET,
  BG_FLAGS_AGG_MOMENTARY_RANGE_OFFSET,
  BG_FLAGS_AGG_SHORTTERM_MEAN_OFFSET,
  BG_FLAGS_AGG_SHORTTERM_MAXIMUM_OFFSET,
  BG_FLAGS_AGG_SHORTTERM_RANGE_OFFSET,
  BG_FLAGS_AGG_SAMPLEPEAK_OFFSET,
  BG_FLAGS_AGG_TRUEPEAK_OFFSET,
  BG_FLAGS_AGG_MAX_OFFSET,
};

enum bg_flags_agg {
  BG_FLAGS_AGG_MOMENTARY_MEAN=1<<BG_FLAGS_AGG_MOMENTARY_MEAN_OFFSET,
  BG_FLAGS_AGG_MOMENTARY_MAXIMUM=1<<BG_FLAGS_AGG_MOMENTARY_MAXIMUM_OFFSET,
  BG_FLAGS_AGG_MOMENTARY_RANGE=1<<BG_FLAGS_AGG_MOMENTARY_RANGE_OFFSET,
  BG_FLAGS_AGG_SHORTTERM_MEAN=1<<BG_FLAGS_AGG_SHORTTERM_MEAN_OFFSET,
  BG_FLAGS_AGG_SHORTTERM_MAXIMUM=1<<BG_FLAGS_AGG_SHORTTERM_MAXIMUM_OFFSET,
  BG_FLAGS_AGG_SHORTTERM_RANGE=1<<BG_FLAGS_AGG_SHORTTERM_RANGE_OFFSET,
  BG_FLAGS_AGG_SAMPLEPEAK=1<<BG_FLAGS_AGG_SAMPLEPEAK_OFFSET,
  BG_FLAGS_AGG_TRUEPEAK=1<<BG_FLAGS_AGG_TRUEPEAK_OFFSET,
  BG_FLAGS_AGG_MAX=1<<BG_FLAGS_AGG_MAX_OFFSET,
  BG_FLAGS_AGG_MOMENTARY
      =BG_FLAGS_AGG_MOMENTARY_MAXIMUM
      |BG_FLAGS_AGG_MOMENTARY_MEAN
      |BG_FLAGS_AGG_MOMENTARY_RANGE,
  BG_FLAGS_AGG_SHORTTERM
      =BG_FLAGS_AGG_SHORTTERM_MAXIMUM
      |BG_FLAGS_AGG_SHORTTERM_MEAN
      |BG_FLAGS_AGG_SHORTTERM_RANGE,
  BG_FLAGS_AGG_PEAK
      =BG_FLAGS_AGG_SAMPLEPEAK
      |BG_FLAGS_AGG_TRUEPEAK,
  BG_FLAGS_AGG_ALL
      =BG_FLAGS_AGG_MOMENTARY
      |BG_FLAGS_AGG_SHORTTERM
      |BG_FLAGS_AGG_SAMPLEPEAK
      |BG_FLAGS_AGG_TRUEPEAK,
  BG_FLAGS_AGG_INTEGRATED=BG_FLAGS_AGG_MOMENTARY_MEAN,
};
#endif // ]

////////
enum bg_flags_ext {
  BG_FLAGS_EXT_RENAME=1<<1,
  BG_FLAGS_EXT_CSV=1<<2,
#if 0 // [
  BG_FLAGS_EXT_JPG=1<<3,
#elif defined (BG_FILE) // ] [
  BG_FLAGS_EXT_COPY=1<<3,
#endif // ]
  BG_FLAGS_EXT_TAGS=1<<4,
  BG_FLAGS_EXT_ALL
      =BG_FLAGS_EXT_RENAME
      |BG_FLAGS_EXT_CSV
#if 0 // [
      |BG_FLAGS_EXT_JPG
#elif defined (BG_FILE) // ] [
      |BG_FLAGS_EXT_COPY
#endif // ]
      |BG_FLAGS_EXT_TAGS,
};

////////
enum bg_flags_mode {
  BG_FLAGS_MODE_APPLY=1<<0,
#if defined (BG_BWF_TAGS) // [
  BG_FLAGS_MODE_TAGS_RG=1<<1,
  BG_FLAGS_MODE_TAGS_BWF=1<<2,
#endif // ]
  BG_FLAGS_MODE_TAGS_TRACK=1<<3,
  BG_FLAGS_MODE_TAGS_ALBUM=1<<4,
  BG_FLAGS_MODE_TAGS_ALL
      =BG_FLAGS_MODE_TAGS_TRACK
      |BG_FLAGS_MODE_TAGS_ALBUM,
};

#if defined (BG_LIB1770) // [
////////
enum bg_flags_norm {
  BG_FLAGS_NORM_NULL=0ul,
  BG_FLAGS_NORM_EBU=1ul<<1ul,
  BG_FLAGS_NORM_ATSC=1ul<<2ul,
  BG_FLAGS_NORM_REPLAYGAIN=1ul<<3ul,
};
#endif // ]

////////
struct bg_param {
#if defined (BG_PRINT) // [
  struct {
    bg_print_vmt_t *vtbl;
  } print;
#endif // ]

  int loglevel;
  int purged;
#if defined (FF_OUTPUT_CALLBACK_CODEC_ID) // [
  struct {
#if defined (_WIN32) // [
    char name[64];
#else // ] [
    char *name;
#endif // ]
  } codec;
#endif // ]
#if defined (BG_TEMP_PREFIX) // [
  const ffchar_t *temp_prefix;
#endif // ]
#if defined (BG_SKIP_FOLDER) // [
  int skip_folder;
#endif // ]
#if defined (_WIN32) // [
  // if LANG is set to e.g. "en_US.UTF-8" we assume we're run from
  // e.g. MSYS2 shell undestanding UTF-8 otherwise from MS console using
  // codepage OEM. In the latter case we need an OEM representation of
  // e.g. basename.
  int oem;
#endif // ]
  ff_param_decode_t decode;

  struct {
    FILE *f;
  } result;

  struct {
    const ffchar_t *dirname;
  } output;

  int ai,vi;

  struct {
    bg_flags_ext_t extension;
    bg_flags_mode_t mode;
#if defined (BG_LIB1770) // [
    bg_flags_agg_t aggregate;
    bg_flags_norm_t norm;
#endif // ]
  } flags;

  int dump;
  int stereo;
#if defined (BG_LIB1770) // [
  double norm;
  double preamp;

  struct {
    int enabled;
    double value;
  } weight;
#endif // ]
  int time;
  int lfe;
#if defined (_WIN32) && defined (BG_WIN32_CREATE_LOCALE) // [
  _locale_t locale;
#endif // ]
#if defined (BG_OVERWRITE) // [
  int overwrite;
#endif // ]
  const ffchar_t *unit;

  struct {
    char pfx[128];
  } tag;

#if defined (FF_OUTPUT_CALLBACK_CODEC_ID) // [
  struct {
    ffchar_t *sfx;
  } out;
#endif // ]

  struct {
    const ffchar_t *audio;
    const ffchar_t *video;
  } ext;

#if defined (BG_INTERVAL) // [
  struct {
    int64_t begin;
    int64_t duration;
  } interval;
#endif // ]

#if defined (BG_LIB1770) // [
  bg_param_block_t momentary;
  bg_param_block_t shortterm;
#endif // ]
};

#if defined (BG_PRINT) // [
///////////////////////////////////////////////////////////////////////////////
struct bg_print_conf {
#if defined (BG_LIB1770) // [
  bg_flags_agg_t aggregate;
#endif // ]

#if defined (_WIN32) // [
  // when not writing to the console/shell (i.e. to a file) we need to use
  // the wide character representation.
  struct {
    const wchar_t *label;

    struct {
      const wchar_t *classic;
      const wchar_t *xml;
    } format;
  } w;
#endif // ]

  // narrow character representation.
  struct {
    const char *label;

    struct {
      const char *classic;
      const char *xml;
    } format;
  } n;

  int argc;
  double (*argv[2])(bg_tree_t *tree);
};

extern bg_print_conf_t bg_print_conf[BG_FLAGS_AGG_MAX_OFFSET];
#endif // ]

#if defined (__cplusplus) // [
}
#endif // ]
#endif // __BG_H__ ]
