/*
 * bgx.c
 * Copyright (C) 2014-2019 Peter Belkner <pbelkner@users.sf.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2.0 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 */
/*
 * once meant as an example program on how to deal with libbg.a (and in
 * turn with libff.a) finally serving as bs1770gain's implementation.
 */
#if defined (_WIN32) // [
#include <fcntl.h>
#include <getoptW.h>
#else // ] [
#include <ctype.h>
#include <getopt.h>
#endif // ]
#include <locale.h>
#include <bg.h>

///////////////////////////////////////////////////////////////////////////////
#if 1 // [
#if 1 // [
#define BS1770GAIN_TEAM_WHITE \
    FFL(" #TeamWhite")
#else // ] [
#define BS1770GAIN_TEAM_WHITE \
    FFL("")
#endif // ]
#define BS1770GAIN_NANOS_GIGANTUM_HUMERIS_INSIDENTES \
    FFL("Nanos gigantum humeris insidentes.") BS1770GAIN_TEAM_WHITE FFL("\n")
#endif // ]

///////////////////////////////////////////////////////////////////////////////
static void usage(const ffchar_t *path, FILE *f)
{
  path=FFBASENAME(path);
  FFFPUTS(path,f);
#if defined (PACKAGE_VERSION) // [
  FFFPUTC(FFL(' '),f);
  FFFPUTS(FFSTR(PACKAGE_VERSION),f);
#endif // ]
  FFFPUTS(FFL(", Copyright (C) Peter Belkner 2014-2019.\n"),f),
#if defined (BS1770GAIN_NANOS_GIGANTUM_HUMERIS_INSIDENTES) // [
  FFFPUTS(BS1770GAIN_NANOS_GIGANTUM_HUMERIS_INSIDENTES,f);
#endif // ]
#if defined (PACKAGE_URL) // [
  FFFPUTS(FFSTR(PACKAGE_URL),f);
  FFFPUTC(FFL('\n'),f);
#endif // ]
  FFFPUTC(FFL('\n'),f);
  FFFPRINTF(f,FFL("Usage:  %s [options] <file/dir> [<file/dir> ...]\n"),path);
  FFFPUTC(FFL('\n'),f);
  FFFPUTS(FFL("Options:\n"),f);
  FFFPUTS(FFL(" -h,--help:  print this message and exit\n"),f);
  /////////////////////////////////////////////////////////////////////////////
  FFFPUTS(FFL(" -i,--integrated:  calculate integrated loudness\n"),f);
  FFFPUTS(FFL(" -s,--shortterm:  calculate maximum shortterm loudness\n"),f);
  FFFPUTS(FFL(" -m,--momentary:  calculate maximum momentary loudness\n"),f);
  FFFPUTS(FFL(" -r,--range:  calculate loudness range\n"),f);
#if defined (BG_LIB1770) // [
  FFFPUTS(FFL(" -p,--samplepeak:  calculate maximum sample peak\n"),f);
  FFFPUTS(FFL(" -t,--truepeak:  calculate maximum true peak\n"),f);
#endif // ]
#if defined (BG_INTERVAL) // [
  FFFPUTS(FFL(" -b <timestamp>,--begin=<timestamp>:  begin decoding at\n")
      FFL("   timestamp (in microseconds, format: hh:mm:ss.mus)\n"),f);
  FFFPUTS(FFL(" -d <duration>,--duration=<duration>:  let decoding\n")
      FFL("   last duration (in microseconds, format: hh:mm:ss.mus)\n"),f);
#endif // ]
  FFFPUTS(FFL(" -u <method>,--use=<method>:  base replaygain calculation")
          FFL(" on\n")
      FFL("   <method> (with respect to the -a/--apply and -o/--output\n")
#if 0 // [
      FFL("   options. specify either option -o/--output or option\n")
      FFL("   --overwrite but not both.)\n")
      FFL("   available methods:\n")
#else // ] [
      FFL("   options. available methods:\n")
#endif // ]
      FFL("     \"integrated\" (default),\n")
      FFL("     \"momentary\", and\n")
      FFL("     \"shortterm\"\n")
      FFL("   experimental methods:\n")
      FFL("     \"momentary-mean\" (synonym for \"integrated\"),\n")
      FFL("     \"momentary-maximum\" (synonym for \"momentary\"),\n")
      FFL("     \"shortterm-mean\", and\n")
      FFL("     \"shortterm-maximum\" (synonym for \"shortterm\")\n"),f);
  FFFPUTS(FFL(" -a [<weight>]/--apply[=<weight>]:  apply the EBU/ATSC/RG\n")
      FFL("   album gain (in conjunction with the -o/--output option.)\n")
      FFL("   when <weight> out of [0.0 .. 1.0] is provided: album gain\n")
      FFL("   plus <weight>*(track gain - album gain).\n"),f);
  FFFPUTS(FFL(" -o <folder>,--output=<folder>:  write replaygain tags\n")
      FFL("   or apply the EBU/ATSC/RG gain, respectively,\n")
      FFL("   and output to <folder>.\n")
      FFL("   specify either option -o/--output or option --overwrite\n")
      FFL("   but not both.\n"),f);
#if defined (_WIN32) // [
  FFFPUTS(FFL(" -f <file>,--file=<file>:  write analysis to an utf-8 log")
      FFL(" file\n"),f);
  FFFPUTS(FFL(" --utf-16[=<file>]:  write an utf-16 log file instead of")
      FFL (" an utf-8\n")
      FFL("   log file (<file> can be omitted when provided by means of the\n")
      FFL("   -f/--file option.)\n"),f);
#else // ] [
  FFFPUTS(FFL(" -f <file>,--file=<file>:  write analysis to a log file\n"),f);
#endif // ]
#if defined (BG_EXTENSIONS) // [
  FFFPUTS(FFL(" -x,--extensions:  enable following extensions/defaults:\n")
      FFL("   1) rename files according to TITLE tag\n")
      FFL("   2) read metadata from per-folder CSV file \"folder.csv\"\n")
#if 0 // [
      FFL("   3) copy file \"folder.jpg\" from source to destination\n")
      FFL("      folder\n")
#elif defined (BG_FILE) // ] [
      FFL("   3) copy non-audio files from source to destination\n")
      FFL("      folder\n")
#endif // ]
      FFL("   4) automatically add the TRACK and DISC tags\n")
      FFL("   5) convert to stereo\n"),f);
#if defined (BG_LIB1770) // [
  FFFPUTS(FFL("   6) calculate maximum true peak\n"),f);
#endif // ]
#endif // ]
#if 0 && defined (_WIN32) // [
  FFFPUTS(FFL(" -v,--utf-16:  use utf-16 for classical (i.e. non-xml)\n")
      FFL("   output (may be useful for languages containing non-ascii")
          FFL(" symbols,\n")
      FFL("   currently supresses displaying progress of analysis)\n"),f);
#endif // ]
#if defined (BG_LIST) // [
  FFFPUTS(FFL(" -l,--list:  print FFmpeg format/stream information\n"),f);
#endif // ]
  /////////////////////////////////////////////////////////////////////////////
  FFFPUTS(FFL(" --ebu:  calculate replay gain according to EBU R128\n")
      FFL("   (-23.0 LUFS, default)\n"),f);
  FFFPUTS(FFL(" --atsc:  calculate replay gain according to ATSC A/85\n")
      FFL("   (-24.0 LUFS)\n"),f);
  FFFPUTS(FFL(" --replaygain:  calculate replay gain according to\n")
      FFL("   ReplayGain 2.0 (-18.0 LUFS)\n"),f);
  FFFPUTS(FFL(" --track-tags:  write track tags\n"),f);
  FFFPUTS(FFL(" --album-tags:  write album tags\n"),f);
  FFFPUTS(FFL(" --tag-prefix=<prefix>:  instead of \"REPLAYGAIN\",\n")
      FFL("   use <prefix> as replaygain tag prefix\n"),f);
  FFFPUTS(FFL(" --unit=<unit>:  write tags with <unit>\n"),f);
#if 0 // [
  FFFPUTS(FFL(" --apply[=<weight>]:  apply the EBU/ATSC/RG gain to the"
      " output\n")
      FFL("   (in conjunction with option -o/--output),\n")
      FFL("   weight=0.0 (album gain) ... 1.0 (track gain),\n")
      FFL("   default:  0.0 (album gain)\n"),f);
#endif // ]
#if defined (BG_LIST) // [
  FFFPUTS(FFL(" --audio=<index>/--ai=<index>:  select audio index."),f);
#else // ] [
  FFFPUTS(FFL(" --audio=<index>/--ai=<index>:  select audio index")
      FFL(" (corresponds\n")
      FFL("   to [0:<index>] in FFmpeg listing, cf. -l/--list option)\n"),f);
#endif // ]
  FFFPUTS(FFL(" --video=<index>/--vi=<index>:  select video index")
      FFL(" (corresponds\n")
      FFL("   to [0:<index>] in FFmpeg listing, cf. -l/--list option)\n"),f);
  FFFPUTS(FFL(" --stereo:  convert to stereo\n"),f);
  FFFPUTS(FFL(" --drc=<float>:  set AC3 dynamic range compression (DRC)\n"),f);
#if defined (BG_EXTENSIONS) // [
  FFFPUTS(FFL(" --extension=<extension>:  enable extension out of\n")
      FFL("    \"rename\":  rename files according to TITLE tag\n")
      FFL("    \"csv\":  read metadata from per-folder CSV file")
          FFL(" \"folder.csv\"\n")
#if 0 // [
      FFL("    \"jpg\":  copy file \"folder.jpg\" from source to destination\n")
      FFL("       folder\n")
#elif defined (BG_FILE) // ] [
      FFL("    \"copy\":  copy non-audio files from source to destination\n")
      FFL("      folder\n")
      FFL(" folder\n")
#endif // ]
#else // ] [
  FFFPUTS(FFL(" -x [<extension>]?, --extension[=<extension>]?:  enable")
      FFL(" extension\n")
      FFL("   out of\n")
      FFL("    \"rename\":  rename files according to TITLE tag\n")
      FFL("    \"csv\":  read metadata from per-folder CSV file")
          FFL(" \"folder.csv\"\n")
#if 0 // [
      FFL("    \"jpg\":  copy file \"folder.jpg\" from source to destination\n")
      FFL("       folder\n")
#elif defined (BG_FILE) // ] [
      FFL("    \"copy\":  copy non-audio files from source to destination\n")
      FFL("      folder\n")
#endif // ]
#endif // ]
      FFL("    \"tags\":  automatically add the TRACK and DISC tags\n")
      FFL("    \"all\":  all of the above (can be omitted)\n"),f);
#if defined (FF_OUTPUT_CALLBACK_CODEC_ID) // [
  FFFPUTS(FFL(" --suffix=<suffix>:  output to <basename>.<suffix>\n")
      FFL("   (only useful in conjunction with option -o/--output))\n"),f);
#endif // ]
#if 0 // [
  FFFPUTS(FFL(" --format=<format>:  convert to format\n"),f);
#endif // ]
  FFFPUTS(FFL(" --loglevel=<level>:  set FFmpeg loglevel to <level> out of\n")
      FFL("    \"quiet\",\n")
      FFL("    \"panic\",\n")
      FFL("    \"fatal\",\n")
      FFL("    \"error\",\n")
      FFL("    \"warning\",\n")
      FFL("    \"info\",\n")
      FFL("    \"verbose\", or\n")
      FFL("    \"debug\"\n"),f);
#if defined (BG_PRINT) // [
  FFFPUTS(FFL(" --xml:  print results in xml format\n"),f);
#endif // ]
  FFFPUTS(FFL(" --time:  print out duration of program invocation\n"),f);
  FFFPUTS(FFL(" --norm=<float>:  norm loudness to float.\n"),f);
#if 0 // [
  FFFPUTS(FFL(" --preamp=<preamp>:\n"),f);
  FFFPUTS(FFL(" --stero:\n"),f);
  FFFPUTS(FFL(" --rg-tags:\n"),f);
  FFFPUTS(FFL(" --bwf-tags:\n"),f);
#endif // ]
#if defined (BG_OVERWRITE) // [
#if 0 // [
  FFFPUTS(FFL(" --overwrite:  replace any input files by their respective\n")
      FFL("   output files (this option takes precedence over -o/--output).")
          FFL(" An\n")
      FFL("   input file is not really replaced but moved to a file with")
          FFL(" a name\n")
      FFL("   constructed from the old name by")
          FFL(" \"<old name>.BS1770GAIN\".\n"),f);
#else // ] [
  FFFPUTS(FFL(" --overwrite:  replace any source file by its respective\n")
      FFL("    target file. specify either option -o/--output or option\n")
      FFL("    --overwrite but not both.\n")
      FFL("    WARNING:  use this option with extrem care. The source\n")
      FFL("      files will definitely be lost!!! For this reason using\n")
      FFL("      this option is discouraged. It's provided just for\n")
      FFL("      completeness.\n"),f);
#endif // ]
#endif // ]
#if defined (BG_CHANNEL_LFE) // [
  FFFPRINTF(f,FFL(" --lfe=<lfe channel>  overwrite the default <lfe")
      FFL(" channel>\n")
      FFL("    (default: %d.)\n"),BG_CHANNEL_LFE);
#endif // ]
#if defined (FF_OUTPUT_CALLBACK_CODEC_ID) // [
  /*
  FFFPUTS(FFL(" --suffix=<name>:  use suffix \"<name>\" for")
      FFL(" output file.\n"),f);
  */
  FFFPUTS(FFL(" --codec=<name>:  use audio codec \"<name>\" for output\n")
      FFL("    default: \"flac\".)\n"),f);
#endif // ]
#if defined (BG_TEMP_PREFIX) // [
  FFFPUTS(FFL(" --temp-prefix=<prefix>:  create temporary files with\n")
      FFL("    prefix \"<prefix>\" default: \"")
      BG_TEMP_PREFIX
      FFL("\".)\n"),f);
#endif // ]
#if defined (BG_SKIP_FOLDER) // [
  FFFPUTS(FFL(" --skip-folder:  suppress printing folder results.\n"),f);
#endif // ]
  /////////////////////////////////////////////////////////////////////////////
  FFFPUTC(FFL('\n'),f);
  FFFPUTS(FFL("Experimental options:\n"),f);
  ////////
  FFFPUTS(FFL("1) momentary block\n"),f);
  FFFPUTS(FFL(" --momentary-mean:  calculate mean loudness based on\n")
      FFL("   momentary block (same as --integrated)\n"),f);
  FFFPUTS(FFL(" --momentary-maximum:  calculate maximum loudness based\n")
      FFL("   on momentary block (same as --momentary)\n"),f);
  FFFPUTS(FFL(" --momentary-range:  calculate loudness range based on\n")
      FFL("   momentary block\n"),f);
  FFFPUTS(FFL(" --momentary-length=<ms>:  length of momentary block\n")
      FFL("   in milliseconds (default: 400)\n"),f);
  FFFPUTS(FFL(" --momentary-overlap=<percent>:  overlap of momentary\n")
      FFL("   block in percent (default: 75)\n"),f);
  FFFPUTS(FFL(" --momentary-mean-gate=<gate>:  silence gate for mean\n")
      FFL("   measurement of momentary block (default: -10.0)\n"),f);
  FFFPUTS(FFL(" --momentary-range-gate=<gate>:  silence gate for range\n")
      FFL("   measurement of momentary block (default: -20.0)\n"),f);
  FFFPUTS(FFL(" --momentary-range-lower-bound=<float>:  lower bound for\n")
      FFL("   range measurement of momentary block (default: 0.1)\n"),f);
  FFFPUTS(FFL(" --momentary-range-upper-bound=<float>:  upper bound for\n")
      FFL("   range measurement of momentary block (default: 0.95)\n"),f);
  ////////
  FFFPUTS(FFL("2) shortterm block\n"),f);
  FFFPUTS(FFL(" --shortterm-mean:  calculate mean loudness based on\n")
      FFL("   shortterm block\n"),f);
  FFFPUTS(FFL(" --shortterm-maximum:  calculate maximum loudness based\n")
      FFL("   on shortterm block (same as --shortterm)\n"),f);
  FFFPUTS(FFL(" --shortterm-range:  calculate loudness range based on\n")
      FFL("   shortterm block (same as --range)\n"),f);
  FFFPUTS(FFL(" --shortterm-length=<ms>:  length of shortterm block\n")
      FFL("   in milliseconds (default: 3000)\n"),f);
  FFFPUTS(FFL(" --shortterm-overlap <percent>:  overlap of shortterm\n")
      FFL("   block in percent (default: 67)\n"),f);
  FFFPUTS(FFL(" --shortterm-mean-gate=<gate>:  silence gate for mean\n")
      FFL("   measurement of shortterm block (default: -10.0)\n"),f);
  FFFPUTS(FFL(" --shortterm-range-gate=<gate>:  silence gate for range\n")
      FFL("   measurement of shortterm block (default: -20.0)\n"),f);
  FFFPUTS(FFL(" --shortterm-range-lower-bound=<float>:  lower bound for\n")
      FFL("   range measurement of shortterm block (default: 0.1)\n"),f);
  FFFPUTS(FFL(" --shortterm-range-upper-bound=<float>:  upper bound for\n")
      FFL("   range measurement of shortterm block (default: 0.95)\n"),f);
  /////////////////////////////////////////////////////////////////////////////
  FFFPUTS(FFL("\n"),f);
  FFFPUTS(FFL("Command line arguments may appear in any order.\n"),f);
}

///////////////////////////////////////////////////////////////////////////////
// https://stackoverflow.com/questions/190543/how-can-i-change-the-width-of-a-windows-console-window
// https://docs.microsoft.com/en-us/windows/console/window-and-screen-buffer-size
// David Tran: Logging Function Entry and Exit in C
//  https://davidtranscend.com/blog/log-function-entry-exit-c/
#if defined (_WIN32) // [
int wmain(int argc, wchar_t *const *argv)
#else // ] [
int main(int argc, char *const *argv)
#endif // ]
{
  enum {
#if 0 // [
    BG_ARG_APPLY=FFL('z')+1,
    BG_ARG_UNIT,
#else // ] [
    BG_ARG_UNIT=FFL('z')+1,
#endif // ]
#if defined (_WIN32) // [
    BG_ARG_UTF_16,
#endif // ]
    BG_ARG_DRC,
#if ! defined (FF_OUTPUT_CALLBACK_CODEC_ID) // [
    BG_ARG_FORMAT,
#endif // ]
    BG_ARG_LOGLEVEL,
#if defined (BG_LIB1770) // [
    BG_ARG_NORM,
    BG_ARG_PREAMP,
#endif // ]
    BG_ARG_TIME,
    BG_ARG_EBU,
#if defined (BG_EXTENSIONS) // [
    BG_ARG_EXTENSION,
#endif // ]
#if defined (FF_OUTPUT_CALLBACK_CODEC_ID) // [
    BG_ARG_SUFFIX,
#endif // ]
    BG_ARG_ATSC,
    BG_ARG_AUDIO,
    BG_ARG_VIDEO,
    BG_ARG_STEREO,
    BG_ARG_REPLAYGAIN,
    BG_ARG_RG_TAGS,
#if defined (BG_BWF_TAGS) // [
    BG_ARG_BWF_TAGS,
#endif // ]
    BG_ARG_TAGS_TRACK,
    BG_ARG_TAGS_ALBUM,
#if defined (BG_PRINT) // [
    BG_ARG_XML,
#endif // ]
    BG_ARG_TAG_PREFIX,
#if defined (BG_OVERWRITE) // [
    BG_ARG_OVERWRITE,
#endif // ]
    BG_ARG_CH_LFE,
#if defined (FF_OUTPUT_CALLBACK_CODEC_ID) // [
    //BG_ARG_OUTPUT_SUFFIX,
    BG_ARG_OUTPUT_CODEC,
#endif // ]
#if defined (BG_TEMP_PREFIX) // [
    BG_ARG_TEMP_PREFIX,
#endif // ]
#if defined (BG_SKIP_FOLDER) // [
    BG_ARG_SKIP_FOLDER,
#endif // ]
#if defined (BG_LIB1770) // [
    ////////////////
#if 0 // [
    BG_ARG_MOMENTARY_MEAN,
    BG_ARG_MOMENTARY_MAXIMUM,
#endif // ]
    BG_ARG_MOMENTARY_RANGE,
    BG_ARG_MOMENTARY_LENGTH,
    BG_ARG_MOMENTARY_OVERLAP,
    BG_ARG_MOMENTARY_MEAN_GATE,
    BG_ARG_MOMENTARY_RANGE_GATE,
    BG_ARG_MOMENTARY_RANGE_LOWER_BOUND,
    BG_ARG_MOMENTARY_RANGE_UPPER_BOUND,
    ////////////////
    BG_ARG_SHORTTERM_MEAN,
#if 0 // [
    BG_ARG_SHORTTERM_MAXIMUM,
    BG_ARG_SHORTTERM_RANGE,
#endif // ]
    BG_ARG_SHORTTERM_LENGTH,
    BG_ARG_SHORTTERM_OVERLAP,
    BG_ARG_SHORTTERM_MEAN_GATE,
    BG_ARG_SHORTTERM_RANGE_GATE,
    BG_ARG_SHORTTERM_RANGE_LOWER_BOUND,
    BG_ARG_SHORTTERM_RANGE_UPPER_BOUND,
#endif // ]
  };

#if defined (HAVE_FF_DYNLOAD) // [
  static const ffchar_t TOOLS[]=FFL("bs1770gain-tools");
#endif // ]
  static const ffchar_t FLAGS[]=
#if defined (BG_LIB1770) // [
      FFL("u:ptismr")
#endif // ]
#if defined (_WIN32) // [
      FFL("v")
#endif // ]
#if defined (BG_INTERVAL) // [
      FFL("b:d:")
#endif // ]
//#if ! defined (BG_EXTENSIONS) // [
//      FFL("o:f:ahlx:");
//#else // ] [
      FFL("o:f:ahlx");
//#endif // ]

  static struct option OPTS[]={
    { FFL("help"),no_argument,NULL,
        FFL('h') },
    { FFL("list"),no_argument,NULL,
        FFL('l') },
#if defined (BG_INTERVAL) // [
    { FFL("begin"),required_argument,NULL,
        FFL('b') },
    { FFL("duration"),required_argument,NULL,
        FFL('d') },
#endif // ]
    { FFL("file"),required_argument,NULL,
        FFL('f') },
    { FFL("utf-8"),required_argument,NULL,
        FFL('f') },
    { FFL("output"),required_argument,NULL,
        FFL('o') },
#if defined (BG_LIB1770) // [
    { FFL("use"),required_argument,NULL,
        FFL('u') },
    { FFL("integrated"),no_argument,NULL,
        FFL('i') },
    { FFL("momentary"),no_argument,NULL,
        FFL('m') },
    { FFL("range"),no_argument,NULL,
        FFL('r') },
    { FFL("samplepeak"),no_argument,NULL,
        FFL('p') },
    { FFL("shortterm"),no_argument,NULL,
        FFL('s') },
    { FFL("truepeak"),no_argument,NULL,
        FFL('t') },
#endif // ]
#if defined (BG_EXTENSIONS) // [
    { FFL("extensions"),no_argument,NULL,
        FFL('x') },
#endif // ]
    { FFL("apply"),optional_argument,NULL,
        FFL('a') },
#if defined (_WIN32) // [
    { FFL("utf-16"),optional_argument,NULL,
        BG_ARG_UTF_16 },
#endif // ]
    { FFL("unit"),required_argument,NULL,
        BG_ARG_UNIT },
    { FFL("audio"),required_argument,NULL,
        BG_ARG_AUDIO },
    { FFL("ai"),required_argument,NULL,
        BG_ARG_AUDIO },
    { FFL("drc"),required_argument,NULL,
        BG_ARG_DRC },
#if defined (BG_EXTENSIONS) // [
    { FFL("extension"),required_argument,NULL,
        BG_ARG_EXTENSION },
#else // ] [
    { FFL("extension"),no_argument,NULL,
        FFL('x') },
#endif // ]
#if defined (FF_OUTPUT_CALLBACK_CODEC_ID) // [
    { FFL("suffix"),required_argument,NULL,
        BG_ARG_SUFFIX },
#endif // ]
#if ! defined (FF_OUTPUT_CALLBACK_CODEC_ID) // [
    { FFL("format"),required_argument,NULL,
        BG_ARG_FORMAT },
#endif // ]
    { FFL("loglevel"),required_argument,NULL,
        BG_ARG_LOGLEVEL },
#if defined (BG_LIB1770) // [
    { FFL("norm"),required_argument,NULL,
        BG_ARG_NORM },
    { FFL("preamp"),required_argument,NULL,
        BG_ARG_PREAMP },
#endif // ]
    { FFL("video"),required_argument,NULL,
        BG_ARG_VIDEO },
    { FFL("vi"),required_argument,NULL,
        BG_ARG_VIDEO },
    { FFL("time"),no_argument,NULL,
        BG_ARG_TIME },
    { FFL("ebu"),no_argument,NULL,
        BG_ARG_EBU },
    { FFL("atsc"),no_argument,NULL,
        BG_ARG_ATSC },
    { FFL("stereo"),no_argument,NULL,
        BG_ARG_STEREO },
    { FFL("replaygain"),no_argument,NULL,
        BG_ARG_REPLAYGAIN },
#if defined (BG_BWF_TAGS) // [
    { FFL("rg-tags"),no_argument,NULL,
        BG_ARG_TAGS_RG },
    { FFL("bwf-tags"),no_argument,NULL,
        BG_ARG_TAGS_BWF },
#endif // ]
    { FFL("track-tags"),no_argument,NULL,
        BG_ARG_TAGS_TRACK },
    { FFL("album-tags"),no_argument,NULL,
        BG_ARG_TAGS_ALBUM },
#if defined (BG_LIB1770) // [
    { FFL("xml"),no_argument,NULL,
        BG_ARG_XML },
#endif // ]
    { FFL("tag-prefix"),required_argument,NULL,
        BG_ARG_TAG_PREFIX },
#if defined (BG_OVERWRITE) // [
    { FFL("overwrite"),no_argument,NULL,
        BG_ARG_OVERWRITE },
#endif // ]
    { FFL("lfe"),required_argument,NULL,
        BG_ARG_CH_LFE },
#if defined (FF_OUTPUT_CALLBACK_CODEC_ID) // [
    /*
    { FFL("suffix"),required_argument,NULL,
        BG_ARG_OUTPUT_SUFFIX },
    */
    { FFL("codec"),required_argument,NULL,
        BG_ARG_OUTPUT_CODEC },
#endif // ]
#if defined (BG_TEMP_PREFIX) // [
    { FFL("temp-prefix"),required_argument,NULL,
        BG_ARG_TEMP_PREFIX },
#endif // ]
#if defined (BG_SKIP_FOLDER) // [
    { FFL("skip-folder"),no_argument,NULL,
        BG_ARG_SKIP_FOLDER },
#endif // ]
#if defined (BG_LIB1770) // [
    /// momentary /////////////////////////////////////////////////////////////
    { FFL("momentary-mean"),no_argument,NULL,
        FFL('i') },
    { FFL("momentary-maximum"),no_argument,NULL,
        FFL('m') },
    { FFL("momentary-range"),no_argument,NULL,
        BG_ARG_MOMENTARY_RANGE },
    { FFL("momentary-length"),required_argument,NULL,
        BG_ARG_MOMENTARY_LENGTH },
    { FFL("momentary-overlap"),required_argument,NULL,
        BG_ARG_MOMENTARY_OVERLAP },
    { FFL("momentary-mean-gate"),required_argument,NULL,
        BG_ARG_MOMENTARY_MEAN_GATE },
    { FFL("momentary-range-gate"),required_argument,NULL,
        BG_ARG_MOMENTARY_RANGE_GATE },
    { FFL("momentary-range-lower-bound"),required_argument,NULL,
        BG_ARG_MOMENTARY_RANGE_LOWER_BOUND },
    { FFL("momentary-range-upper-bound"),required_argument,NULL,
        BG_ARG_MOMENTARY_RANGE_UPPER_BOUND },
    { FFL("shortterm-mean"),no_argument,NULL,
        BG_ARG_SHORTTERM_MEAN },
    /// shortterm /////////////////////////////////////////////////////////////
    { FFL("shortterm-maximum"),no_argument,NULL,
        FFL('s') },
    { FFL("shortterm-range"),no_argument,NULL,
        FFL('r') },
    { FFL("shortterm-length"),required_argument,NULL,
        BG_ARG_SHORTTERM_LENGTH },
    { FFL("shortterm-overlap"),required_argument,NULL,
        BG_ARG_SHORTTERM_OVERLAP },
    { FFL("shortterm-mean-gate"),required_argument,NULL,
        BG_ARG_SHORTTERM_MEAN_GATE },
    { FFL("shortterm-range-gate"),required_argument,NULL,
        BG_ARG_SHORTTERM_RANGE_GATE },
    { FFL("shortterm-range-lower-bound"),required_argument,NULL,
        BG_ARG_SHORTTERM_RANGE_LOWER_BOUND },
    { FFL("shortterm-range-upper-bound"),required_argument,NULL,
        BG_ARG_SHORTTERM_RANGE_UPPER_BOUND },
#endif // ]
    { 0 }
  };

  int code=1;
#if defined (_WIN32) // [
  int utf16=0;
  // if LANG is set to e.g. "en_US.UTF-8" we assume we're run from
  // e.g. MSYS2 shell undestanding UTF-8 otherwise from MS console using
  // codepage OEM.
  const char *lang=getenv("LANG");
#else // ] [
  const char *locale=NULL;
#endif // ]
  const ffchar_t *fpath=NULL;
#if defined (FF_ZERO_INITIALIZER) // [
  bg_param_t param={ 0 };
  bg_visitor_t analyzer={ 0 };
#if defined (BG_ANNOTATOR) // [
  bg_visitor_t annotator={ 0 };
#endif // ]
  bg_visitor_t muxer={ 0 };
  bg_tree_t root={ 0 };
#else // ] [
  bg_param_t param;
  bg_visitor_t analyzer;
#if defined (BG_ANNOTATOR) // [
  bg_visitor_t annotator;
#endif // ]
  bg_visitor_t muxer;
  bg_tree_t root;
#endif // ]
  int c;
#if defined (BG_LIB1770) // [
  double overlap;
#endif // ]
  unsigned version;
#if 0 && defined (FF_OUTPUT_CALLBACK_CODEC_ID) // [
#if defined (_WIN32) // [
  char name[64];
#endif // ]
  AVCodec *codec;
#endif // ]

#if ! defined (FF_ZERO_INITIALIZER) // [
  /////////////////////////////////////////////////////////////////////////////
  memset(&param,0,sizeof param);
  memset(&analyzer,0,sizeof analyzer);
#if defined (BG_ANNOTATOR) // [
  memset(&annotator,0,sizeof annotator);
#endif // ]
  memset(&muxer,0,sizeof muxer);
  memset(&root,0,sizeof root);
#endif // ]

  /////////////////////////////////////////////////////////////////////////////
  if (argc<2) {
    FFMESSAGE(FFL("missing command line arguments"));
    FFFPUTC(FFL('\n'),stderr);
    usage(argv[0],stderr);
    goto e_arg;
  }

  /////////////////////////////////////////////////////////////////////////////
#if defined (BG_PRINT) // [
  param.print.vtbl=&bg_print_classic_vmt;
#endif // ]
#if 0 // [
  param.result.f=stderr;
#else // ] [
  param.result.f=stdout;
#endif // ]
  param.decode.request.sample_fmt=-1;
  param.decode.request.channel_layout=-1;
  param.decode.drc.enabled=0;
  param.decode.drc.scale=0.0;
  param.ai=-1;
  param.vi=-1;
  param.loglevel=AV_LOG_QUIET;
#if defined (BG_TEMP_PREFIX) // [
  param.temp_prefix=BG_TEMP_PREFIX;
#endif // ]
  param.dump=0;
  param.stereo=0;
#if defined (BG_LIB1770) // [
  param.flags.norm=BG_FLAGS_NORM_EBU;
  param.norm=-23.0;
  param.preamp=0.0;
  param.weight.enabled=0;
  param.weight.value=1.0;
#endif // ]
#if defined (LIB1770_LFE) // [
  param.lfe=LIB1770_LFE;
#elif defined (BG_CHANNEL_LFE) // [
  param.lfe=BG_CHANNEL_LFE;
#else // ] [
  param.lfe=-1;
#endif // ]
  param.unit=FFL("LU");
  strcpy(param.tag.pfx,"REPLAYGAIN_");
  param.ext.audio=FFL("flac");
  param.ext.video=FFL("mkv");
#if defined (BG_INTERVAL) // [
  param.interval.begin=-1ll;
  param.interval.duration=-1ll;
#endif // ]
#if defined (BG_LIB1770) // [
#if 0 // [
  param.flags.aggregate|=BG_FLAGS_AGG_ALL;
#else // ] [
  param.flags.aggregate|=BG_FLAGS_AGG_INTEGRATED;
#endif // ]
  ////////
  param.momentary.ms=400.0;
  param.momentary.partition=4;
  param.momentary.mean.gate=-10.0;
  param.momentary.range.gate=-20.0;
  param.momentary.range.lower_bound=0.1;
  param.momentary.range.upper_bound=0.95;
  ////////
  param.shortterm.ms=3000.0;
  param.shortterm.partition=3;
  param.shortterm.mean.gate=-10.0;
  param.shortterm.range.gate=-20.0;
  param.shortterm.range.lower_bound=0.1;
  param.shortterm.range.upper_bound=0.95;
#endif // ]

  /////////////////////////////////////////////////////////////////////////////
  while (0<=(c=FF_GETOPT_LONG(argc,argv,FLAGS,OPTS,NULL))) {
    switch (c) {
    ///////////////////////////////////////////////////////////////////////////
    case FFL('?'):
      FFFPUTC(FFL('\n'),stderr);
      usage(argv[0],stderr);
      goto e_arg;
    ///////////////////////////////////////////////////////////////////////////
#if defined (BG_INTERVAL) // [
    case FFL('b'):
      param.interval.begin=bg_parse_time(optarg);
      break;
    case FFL('d'):
      param.interval.duration=bg_parse_time(optarg);
      break;
#endif // ]
    case FFL('f'):
      fpath=optarg;
      break;
    case FFL('o'):
      if (!optarg) {
        DMESSAGE("missing argument to option -o/--output");
        goto e_arg;
      }
#if 0 // [
      else if (bgmkdir(optarg)<0) {
        DVMESSAGE("creating directory \"%s\"",optarg);
        goto e_arg;
      }
#endif // ]
#if defined (BG_OVERWRITE) // [
      else if (param.overwrite) {
        DMESSAGE("specify either option -o/--output or option --overwrite"
            " but not both");
        usage(argv[0],stderr);
        goto e_arg;
      }
#endif // ]

      param.output.dirname=bg_pathnorm(optarg);
      break;
#if defined (BG_LIB1770) // [
    case FFL('u'):
      if (!FFSTRCASECMP(FFL("integrated"),optarg)
          ||!FFSTRCASECMP(FFL("momentary-mean"),optarg))
        param.flags.aggregate|=BG_FLAGS_AGG_MOMENTARY_MEAN;
      else if (!FFSTRCASECMP(FFL("momentary"),optarg)
          ||!FFSTRCASECMP(FFL("momentary-maximum"),optarg))
        param.flags.aggregate|=BG_FLAGS_AGG_MOMENTARY_MAXIMUM;
      else if (!FFSTRCASECMP(FFL("shortterm-mean"),optarg))
        param.flags.aggregate|=BG_FLAGS_AGG_SHORTTERM_MEAN;
      else if (!FFSTRCASECMP(FFL("shortterm"),optarg)
          ||!FFSTRCASECMP(FFL("shortterm-maximum"),optarg))
        param.flags.aggregate|=BG_FLAGS_AGG_SHORTTERM_MAXIMUM;
      else {
        FFFPRINTF(stderr,FFL("Error: method \"%s\" not recognized.\n\n"),
            optarg);
        usage(argv[0],stderr);
        goto e_arg;
      }

      break;
#endif // ]
    /// without argument ////////////////////////////////////////////////////
    case FFL('a'):
      param.flags.mode|=BG_FLAGS_MODE_APPLY;

      if (optarg) {
        param.weight.enabled=1;
        param.weight.value=FFATOF(optarg);

        if (param.weight.value<0.0||1.0<param.weight.value) {
          DMESSAGE("weight out of range");
          usage(argv[0],stderr);
          goto e_arg;
        }
      }

      break;
    case FFL('h'):
      code=0;
      usage(argv[0],stdout);
      goto e_arg;
#if defined (BG_LIST) // [
    case FFL('l'):
      param.result.f=stderr;
      param.dump=1;
      break;
#endif // ]
#if defined (BG_LIB1770) // [
    case FFL('i'):
      param.flags.aggregate|=BG_FLAGS_AGG_MOMENTARY_MEAN;
      break;
    case FFL('s'):
      param.flags.aggregate|=BG_FLAGS_AGG_SHORTTERM_MAXIMUM;
      break;
    case FFL('m'):
      param.flags.aggregate|=BG_FLAGS_AGG_MOMENTARY_MAXIMUM;
      break;
    case FFL('r'):
      param.flags.aggregate|=BG_FLAGS_AGG_SHORTTERM_RANGE;
      break;
    case FFL('p'):
      param.flags.aggregate|=BG_FLAGS_AGG_SAMPLEPEAK;
      break;
    case FFL('t'):
      param.flags.aggregate|=BG_FLAGS_AGG_TRUEPEAK;
      break;
#endif // ]
    case FFL('x'):
#if ! defined (BG_EXTENSIONS) // [
      if (!optarg) {
#endif // ]
#if defined (BG_LIB1770) // [
        param.flags.aggregate|=BG_FLAGS_AGG_TRUEPEAK;
#endif // ]
        param.flags.extension=BG_FLAGS_EXT_ALL;
        param.stereo=1;
        param.decode.drc.enabled=1;
        param.decode.drc.scale=0.0;
#if ! defined (BG_EXTENSIONS) // [
      }
      else if (0==FFSTRCASECMP(FFL("rename"),optarg))
        param.flags.extension|=BG_FLAGS_EXT_RENAME;
      else if (0==FFSTRCASECMP(FFL("csv"),optarg))
        param.flags.extension|=BG_FLAGS_EXT_CSV;
#if 0 // [
      else if (0==FFSTRCASECMP(FFL("jpg"),optarg))
        param.flags.extension|=BG_FLAGS_EXT_JPG;
#elif defined (BG_FILE) // ] [
      else if (0==FFSTRCASECMP(FFL("copy"),optarg))
        param.flags.extension|=BG_FLAGS_EXT_COPY;
#endif // ]
      else if (0==FFSTRCASECMP(FFL("tags"),optarg))
        param.flags.extension|=BG_FLAGS_EXT_TAGS;
      else if (0==FFSTRCASECMP(FFL("all"),optarg))
        param.flags.extension|=BG_FLAGS_EXT_ALL;
      else {
        FFFPRINTF(stderr,FFL("Error: extension \"%s\" not recognized.\n\n"),
            optarg);
        usage(argv[0],stderr);
        goto e_arg;
      }
#endif // ]
      break;
    /// without flag ////////////////////////////////////////////////////////
    case BG_ARG_AUDIO:
      param.ai=FFATOI(optarg);
      break;
    case BG_ARG_VIDEO:
      param.vi=FFATOI(optarg);
      break;
    case BG_ARG_UNIT:
      param.unit=optarg;
      break;
#if defined (_WIN32) // [
    case BG_ARG_UTF_16:
      utf16=1;

      if (optarg)
        fpath=optarg;

      break;
#endif // ]
    case BG_ARG_DRC:
      param.decode.drc.enabled=1;
      param.decode.drc.scale=FFATOF(optarg);
      break;
#if defined (BG_EXTENSIONS) // [
    case BG_ARG_EXTENSION:
      if (0==FFSTRCASECMP(FFL("rename"),optarg))
        param.flags.extension|=BG_FLAGS_EXT_RENAME;
      else if (0==FFSTRCASECMP(FFL("csv"),optarg))
        param.flags.extension|=BG_FLAGS_EXT_CSV;
#if 0 // [
      else if (0==FFSTRCASECMP(FFL("jpg"),optarg))
        param.flags.extension|=BG_FLAGS_EXT_JPG;
#elif defined (BG_FILE) // ] [
      else if (0==FFSTRCASECMP(FFL("copy"),optarg))
        param.flags.extension|=BG_FLAGS_EXT_COPY;
#endif // ]
      else if (0==FFSTRCASECMP(FFL("tags"),optarg))
        param.flags.extension|=BG_FLAGS_EXT_TAGS;
      else if (0==FFSTRCASECMP(FFL("all"),optarg))
        param.flags.extension|=BG_FLAGS_EXT_ALL;
      else {
        FFFPRINTF(stderr,FFL("Error: extension \"%s\" not recognized.\n\n"),
            optarg);
        usage(argv[0],stderr);
        goto e_arg;
      }
 
      break;
#endif // ]
#if defined (FF_OUTPUT_CALLBACK_CODEC_ID) // [
    case BG_ARG_SUFFIX:
      param.out.sfx=optarg;
      break;
#endif // ]
    case BG_ARG_TAG_PREFIX:
#if defined (_WIN32) // [
      ff_wcs2str(optarg,param.tag.pfx,CP_UTF8,(sizeof param.tag.pfx)-2);
#else // ] [
      strncpy(param.tag.pfx,optarg,(sizeof param.tag.pfx)-2);
#endif // ]
      param.tag.pfx[strlen(param.tag.pfx)+1]='\0';
      param.tag.pfx[strlen(param.tag.pfx)]='_';
      break;
#if defined (BG_OVERWRITE) // [
    case BG_ARG_OVERWRITE:
      if (param.output.dirname) {
        DMESSAGE("specify either option -o/--output or option --overwrite"
            " but not both");
        usage(argv[0],stderr);
        goto e_arg;
      }

      param.overwrite=1;
      break;
#endif // ]
    case BG_ARG_CH_LFE:
      param.lfe=FFATOI(optarg);
      break;
#if defined (FF_OUTPUT_CALLBACK_CODEC_ID) // [
    case BG_ARG_OUTPUT_CODEC:
#if defined (_WIN32) // [
      ff_wcs2str(optarg,param.codec.name,CP_UTF8,(sizeof param.codec.name)-1);
#else // ] [
      param.codec.name=optarg;
#endif // ]
      break;
#endif // ]
#if defined (BG_TEMP_PREFIX) // [
    case BG_ARG_TEMP_PREFIX:
      if (!*optarg) {
        DMESSAGE("missing argument to --temp-prefix");
        usage(argv[0],stderr);
        goto e_arg;
      }

      param.temp_prefix=optarg;
      break;
#endif // ]
#if defined (BG_SKIP_FOLDER) // [
    case BG_ARG_SKIP_FOLDER:
      param.skip_folder=1;
      break;
#endif // ]
    case BG_ARG_LOGLEVEL:
      if (0==FFSTRCASECMP(FFL("quiet"),optarg))
        param.loglevel=AV_LOG_QUIET;
      else if (0==FFSTRCASECMP(FFL("panic"),optarg))
        param.loglevel=AV_LOG_PANIC;
      else if (0==FFSTRCASECMP(FFL("fatal"),optarg))
        param.loglevel=AV_LOG_FATAL;
      else if (0==FFSTRCASECMP(FFL("error"),optarg))
        param.loglevel=AV_LOG_ERROR;
      else if (0==FFSTRCASECMP(FFL("warning"),optarg))
        param.loglevel=AV_LOG_WARNING;
      else if (0==FFSTRCASECMP(FFL("info"),optarg))
        param.loglevel=AV_LOG_INFO;
      else if (0==FFSTRCASECMP(FFL("verbose"),optarg))
        param.loglevel=AV_LOG_VERBOSE;
      else if (0==FFSTRCASECMP(FFL("debug"),optarg))
        param.loglevel=AV_LOG_DEBUG;
      else {
        FFFPRINTF(stderr,FFL("Error: loglevel \"%s\" not recognized.\n\n"),
            optarg);
        usage(argv[0],stderr);
        goto e_arg;
      }

      break;
#if defined (BG_LIB1770) // [
    case BG_ARG_NORM:
      param.flags.norm=BG_FLAGS_NORM_NULL;
      param.norm=FFATOF(optarg);
      break;
    case BG_ARG_PREAMP:
      param.preamp=FFATOF(optarg);
      break;
#endif // ]
    /// without flag ////////////////////////////////////////////////////////
#if defined (BG_LIB1770) // [
    case BG_ARG_ATSC:
      param.flags.norm=BG_FLAGS_NORM_ATSC;
      param.norm=-24.0;
      break;
    case BG_ARG_EBU:
      param.flags.norm=BG_FLAGS_NORM_EBU;
      param.norm=-23.0;
      break;
    case BG_ARG_REPLAYGAIN:
      param.flags.norm=BG_FLAGS_NORM_REPLAYGAIN;
      param.norm=-18.0;
      param.unit=FFL("dB");
      break;
#endif // ]
#if defined (BG_BWF_TAGS) // [
    case BG_ARG_TAGS_RG:
      param.flags.mode|=BG_FLAGS_MODE_TAGS_RG;
      break;
    case BG_ARG_TAGS_BWF:
      param.flags.mode|=BG_FLAGS_MODE_TAGS_BWF;
      break;
#endif // ]
    case BG_ARG_TAGS_TRACK:
      param.flags.mode|=BG_FLAGS_MODE_TAGS_TRACK;
      break;
    case BG_ARG_TAGS_ALBUM:
      param.flags.mode|=BG_FLAGS_MODE_TAGS_ALBUM;
      break;
#if defined (BG_PRINT) // [
    case BG_ARG_XML:
      param.print.vtbl=&bg_print_xml_vmt;
      break;
#endif // ]
    case BG_ARG_TIME:
      param.time=1;
      break;
    case BG_ARG_STEREO:
      param.stereo=1;
      break;
#if defined (BG_LIB1770) // [
    /////////////////////////////////////////////////////////////////////////
    case BG_ARG_MOMENTARY_RANGE:
      param.flags.aggregate|=BG_FLAGS_AGG_MOMENTARY_RANGE;
      break;
    ////////
    case BG_ARG_MOMENTARY_LENGTH:
      param.momentary.ms=FFATOF(optarg);
      break;
    case BG_ARG_MOMENTARY_OVERLAP:
      overlap=FFATOF(optarg);

      if (0.0<=overlap&&overlap<100.0)
        param.momentary.partition=floor(100.0/(100.0-overlap)+0.5);
      else {
        FFFPUTS(FFL("Error: overlap out of range [0..100).\n\n"),stderr);
        usage(argv[0],stderr);
        goto e_arg;
      }

      break;
    case BG_ARG_MOMENTARY_MEAN_GATE:
      param.momentary.mean.gate=FFATOF(optarg);
      break;
    case BG_ARG_MOMENTARY_RANGE_GATE:
      param.momentary.range.gate=FFATOF(optarg);
      break;
    case BG_ARG_MOMENTARY_RANGE_LOWER_BOUND:
      param.momentary.range.lower_bound=FFATOF(optarg);
      break;
    case BG_ARG_MOMENTARY_RANGE_UPPER_BOUND:
      param.momentary.range.upper_bound=FFATOF(optarg);
      break;
    /////////////////////////////////////////////////////////////////////////
    case BG_ARG_SHORTTERM_MEAN:
      param.flags.aggregate|=BG_FLAGS_AGG_SHORTTERM_MEAN;
      break;
#if 0 // [
    case BG_ARG_SHORTTERM_RANGE:
      param.flags.aggregate|=BG_FLAGS_AGG_SHORTTERM_RANGE;
      break;
#endif // ]
    ////////
    case BG_ARG_SHORTTERM_LENGTH:
      param.shortterm.ms=FFATOF(optarg);
      break;
    case BG_ARG_SHORTTERM_OVERLAP:
      overlap=FFATOF(optarg);

      if (0.0<=overlap&&overlap<100.0)
        param.shortterm.partition=floor(100.0/(100.0-overlap)+0.5);
      else {
        FFFPUTS(FFL("Error: overlap out of range [0..100).\n\n"),stderr);
        usage(argv[0],stderr);
        goto e_arg;
      }

      break;
    case BG_ARG_SHORTTERM_MEAN_GATE:
      param.shortterm.mean.gate=FFATOF(optarg);
      break;
    case BG_ARG_SHORTTERM_RANGE_GATE:
      param.shortterm.range.gate=FFATOF(optarg);
      break;
    case BG_ARG_SHORTTERM_RANGE_LOWER_BOUND:
      param.shortterm.range.lower_bound=FFATOF(optarg);
      break;
    case BG_ARG_SHORTTERM_RANGE_UPPER_BOUND:
      param.shortterm.range.upper_bound=FFATOF(optarg);
      break;
#endif // ]
    default:
      if ('c'==optopt)
        FFVMESSAGE(FFL("option -%c requires an argument"),optopt);
      else if (isprint(optopt))
        FFVMESSAGE(FFL("unknown option -%c"),optopt);
      else
        FFVMESSAGE(FFL("unknown option character \\x%x"),optopt);

      usage(argv[0],stderr);
      goto e_arg;
    }
  }

  // in case no class of tags is given we provide both, i.e. album and
  // track tags.
  if (!(param.flags.mode&BG_FLAGS_MODE_TAGS_ALL))
    param.flags.mode|=BG_FLAGS_MODE_TAGS_ALL;

#if defined (HAVE_FF_DYNLOAD) // [
  /////////////////////////////////////////////////////////////////////////////
  // load the FFmpeg libraries from "bs1770gain-tools".
  if (ff_dynload(TOOLS)<0) {
    PBU_DMESSAGE("loading shared libraries");
    goto e_dynload;
  }
#endif // ]

  /////////////////////////////////////////////////////////////////////////////
  version=avutil_version();

  if (LIBAVUTIL_VERSION_MAJOR!=AV_VERSION_MAJOR(version)) {
#if defined (HAVE_FF_DYNLOAD) // [
    DVMESSAGE("wrong version of avutil: %d needed, %d loaded from \"%s\"",
        LIBAVUTIL_VERSION_MAJOR,AV_VERSION_MAJOR(version),ff_dynload_path());
#else // ] [
    DVMESSAGE("wrong version of avutil: %d needed, %d linked",
        LIBAVUTIL_VERSION_MAJOR,AV_VERSION_MAJOR(version));
#endif // ]
    goto e_versions;
  }

#if defined (FF_DYNLOAD_REORDER) // [
  ////////
  version=swresample_version();

  if (LIBSWRESAMPLE_VERSION_MAJOR!=AV_VERSION_MAJOR(version)) {
#if defined (HAVE_FF_DYNLOAD) // [
    DVMESSAGE("wrong version of swresample: %d needed, %d loaded from \"%s\"",
        LIBSWRESAMPLE_VERSION_MAJOR,AV_VERSION_MAJOR(version),
        ff_dynload_path());
#else // ] [
    DVMESSAGE("wrong version of swresample: %d needed, %d linked",
        LIBSWRESAMPLE_VERSION_MAJOR,AV_VERSION_MAJOR(version));
#endif // ]
    goto e_versions;
  }
#endif // ]

  ////////
  version=avcodec_version();

  if (LIBAVCODEC_VERSION_MAJOR!=AV_VERSION_MAJOR(version)) {
#if defined (HAVE_FF_DYNLOAD) // [
    DVMESSAGE("wrong version of avcodec: %d needed, %d loaded from \"%s\"",
        LIBAVCODEC_VERSION_MAJOR,AV_VERSION_MAJOR(version),ff_dynload_path());
#else // ] [
    DVMESSAGE("wrong version of avcodec: %d needed, %d linked",
        LIBAVCODEC_VERSION_MAJOR,version);
#endif // ]
    goto e_versions;
  }

  ////////
  version=avformat_version();

  if (LIBAVFORMAT_VERSION_MAJOR!=AV_VERSION_MAJOR(version)) {
#if defined (HAVE_FF_DYNLOAD) // [
    DVMESSAGE("wrong version of avformat: %d needed, %d loaded from \"%s\"",
        LIBAVFORMAT_VERSION_MAJOR,AV_VERSION_MAJOR(version),ff_dynload_path());
#else // ] [
    DVMESSAGE("wrong version of avformat: %d needed, %d linked",
        LIBAVFORMAT_VERSION_MAJOR,AV_VERSION_MAJOR(version));
#endif // ]
    goto e_versions;
  }

#if ! defined (FF_DYNLOAD_REORDER) // [
  ////////
  version=swresample_version();

  if (LIBSWRESAMPLE_VERSION_MAJOR!=AV_VERSION_MAJOR(version)) {
#if defined (HAVE_FF_DYNLOAD) // [
    DVMESSAGE("wrong version of swresample: %d needed, %d loaded from \"%s\"",
        LIBSWRESAMPLE_VERSION_MAJOR,AV_VERSION_MAJOR(version),
        ff_dynload_path());
#else // ] [
    DVMESSAGE("wrong version of swresample: %d needed, %d linked",
        LIBSWRESAMPLE_VERSION_MAJOR,AV_VERSION_MAJOR(version));
#endif // ]
    goto e_versions;
  }
#endif // ]

#if defined (FF_LINK_SWSCALE) // [
  ////////
  version=swscale_version();

  if (LIBSWSCALE_VERSION_MAJOR!=AV_VERSION_MAJOR(version)) {
#if defined (HAVE_FF_DYNLOAD) // [
    DVMESSAGE("wrong version of swresample: %d needed, %d loaded from \"%s\"",
        LIBSWSCALE_VERSION_MAJOR,AV_VERSION_MAJOR(version),ff_dynload_path());
#else // ] [
    DVMESSAGE("wrong version of swresample: %d needed, %d linked",
        LIBSWSCALE_VERSION_MAJOR,AV_VERSION_MAJOR(version));
#endif // ]
    goto e_versions;
  }
#endif // ]

  ////////
  version=postproc_version();

  if (LIBPOSTPROC_VERSION_MAJOR!=AV_VERSION_MAJOR(version)) {
#if defined (HAVE_FF_DYNLOAD) // [
    DVMESSAGE("wrong version of swresample: %d needed, %d loaded from \"%s\"",
        LIBPOSTPROC_VERSION_MAJOR,AV_VERSION_MAJOR(version),ff_dynload_path());
#else // ] [
    DVMESSAGE("wrong version of swresample: %d needed, %d linked",
        LIBPOSTPROC_VERSION_MAJOR,AV_VERSION_MAJOR(version));
#endif // ]
    goto e_versions;
  }

  ////////
  version=avfilter_version();

  if (LIBAVFILTER_VERSION_MAJOR!=AV_VERSION_MAJOR(version)) {
#if defined (HAVE_FF_DYNLOAD) // [
    DVMESSAGE("wrong version of swresample: %d needed, %d loaded from \"%s\"",
        LIBAVFILTER_VERSION_MAJOR,AV_VERSION_MAJOR(version),ff_dynload_path());
#else // ] [
    DVMESSAGE("wrong version of swresample: %d needed, %d linked",
        LIBAVFILTER_VERSION_MAJOR,AV_VERSION_MAJOR(version));
#endif // ]
    goto e_versions;
  }

  /////////////////////////////////////////////////////////////////////////////
#if (LIBAVFORMAT_VERSION_MAJOR<58) // [
  av_register_all();
#endif // ]
#if (LIBAVFILTER_VERSION_MAJOR<7) // [
  avfilter_register_all();
#endif // ]

  /////////////////////////////////////////////////////////////////////////////
#if defined (_WIN32) // [
#if defined (BG_WIN32_CREATE_LOCALE) // [
  param.locale=_create_locale(LC_ALL,"");
#endif // ]
  // if LANG is set to e.g. "en_US.UTF-8" we assume we're run from
  // e.g. MSYS2 shell undestanding UTF-8 otherwise from MS console using
  // codepage OEM. In the latter case we need an OEM representation of
  // e.g. basename.
  param.oem=!lang||!strstr(lang,"UTF-8");
#else // ] [
  // If not already active we set locale to an arbitrary UTF-8.
  locale=setlocale(LC_ALL,NULL);

  if (strstr(locale,"utf")||strstr(locale,"UTF"))
    locale=NULL;
  else
    setlocale(LC_ALL,"en_US.UTF-8");
#endif // ]


  /////////////////////////////////////////////////////////////////////////////
  if (fpath) {
#if defined (_WIN32) // [
    _wremove(fpath);

    if (utf16) {
      param.result.f=_wfopen(fpath,L"wt,ccs=UTF-16LE");
      param.print.vtbl->encoding(&param,16);
    }
    else {
      param.result.f=_wfopen(fpath,L"wt,ccs=UTF-8");
      param.print.vtbl->encoding(&param,8);
    }
#else // ] [
    remove(fpath);
    param.result.f=fopen(fpath,"w");
    param.print.vtbl->encoding(&param,8);
#endif // ]

    if (!param.result.f) {
#if defined (_WIN32) // [
      DVMESSAGEW(L"opening file \"%s\"",fpath);
#else // ] [
      DVMESSAGE("opening file \"%s\"",fpath);
#endif // ]
      goto e_file;
    }
  }
  else if (param.dump)
    param.result.f=stderr;

  /////////////////////////////////////////////////////////////////////////////
  av_log_set_level(param.loglevel);

  /////////////////////////////////////////////////////////////////////////////
  if (optind==argc) {
    fputs("Error: nothing to analyze.\n\n",stderr);
    usage(argv[0],stderr);
    goto e_analyzer;
  }

  /////////////////////////////////////////////////////////////////////////////
  if (bg_analyzer_create(&analyzer)<0) {
    DMESSAGE("creating analyzer");
    goto e_analyzer;
  }

  /////////////////////////////////////////////////////////////////////////////
#if defined (BG_OVERWRITE) // [
  if (param.output.dirname||param.overwrite) {
#else // ] [
  if (param.output.dirname) {
#endif // ]
#if defined (BG_ANNOTATOR) // [
    // NOTE:  the annotation phase is really needed because it may re-organize
    //   the target tree, cf. bg_album_target_purge()!
    if (bg_annotator_create(&annotator)<0) {
      DMESSAGE("creating annotator");
      goto e_annotator;
    }
#endif // ]

    if (bg_muxer_create(&muxer)<0) {
      DMESSAGE("creating muxer");
      goto e_muxer;
    }
#if ! defined (BG_OVERWRITE) // [
  }
#else // ] [
  }
#endif // ]

  /////////////////////////////////////////////////////////////////////////////
  fputs("analyzing ...\n",stdout);
  fflush(stdout);

  if (bg_root_create(&root,&param)<0) {
    DMESSAGE("creating root");
    goto e_root;
  }

  while (optind<argc) {
#if defined (BG_TREE_CREATE_CHILD_WARNING) // [
    if (bg_root_append(&root,bg_pathnorm(argv[optind]))<0)
      BGVWARNING(FFL("appending \"%s\" to root"),argv[optind]);
#else // ] [
    bg_root_append(&root,bg_pathnorm(argv[optind]));
#endif // ]
    ++optind;
  }

  if (analyzer.vtbl&&root.vtbl->accept(&root,&analyzer)<0) {
    DMESSAGE("analyzing phase");
    goto e_analyzer_accept;
  }

  fflush(stdout);

#if defined (BG_ANNOTATOR) // [
  if (annotator.vtbl&&muxer.vtbl) {
    // NOTE:  the annotation phase is really needed because it may re-organize
    //   the target tree, i.e. avoid a sub-tree containing just one child,
    //   cf. bg_album_target_purge()!
    if (root.vtbl->accept(&root,&annotator)<0) {
      DMESSAGE("annotating phase");
      goto e_annotator_accept;
    }

    fflush(stdout);
#else // ] [
  if (muxer.vtbl) {
#endif // ]
    ///////////////////////////////////////////////////////////////////////////
    fputs("remuxing ...\n",stdout);
    fflush(stdout);

    if (root.vtbl->accept(&root,&muxer)<0) {
      DMESSAGE("muxing phase");
      goto e_muxerer_accept;
    }

    fflush(stdout);
#if ! defined (BG_ANNOTATOR) // [
  }
#else // ] [
  }
#endif // ]

  /////////////////////////////////////////////////////////////////////////////
  fputs("done.\n",stdout);
  fflush(stdout);
  code=0;
//cleanup:
e_muxerer_accept:
#if defined (BG_ANNOTATOR) // [
e_annotator_accept:
#endif // ]
e_analyzer_accept:
  root.vtbl->destroy(&root);
e_root:
  if (muxer.vtbl)
    muxer.vtbl->destroy(&muxer);
e_muxer:
#if defined (BG_ANNOTATOR) // [
  if (annotator.vtbl)
    annotator.vtbl->destroy(&annotator);
e_annotator:
#endif // ]
  if (analyzer.vtbl)
    analyzer.vtbl->destroy(&analyzer);
e_analyzer:
  if (fpath)
    fclose(param.result.f);
e_file:
#if defined (_WIN32) // [
#if defined (BG_WIN32_CREATE_LOCALE) // [
  _free_locale(param.locale);
#endif // ]
#else // ] [
  if (locale)
    setlocale(LC_ALL,locale);
#endif // ]
e_versions:
#if defined (HAVE_FF_DYNLOAD) // [
  ff_unload();
e_dynload:
#endif // ]
e_arg:
  return code;
}
