/*
 * ff_fcprintf.c
 * Copyright (C) 2014-2019 Peter Belkner <pbelkner@users.sf.net>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public
 * License as published by the Free Software Foundation; either
 * version 2.0 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 * MA  02110-1301  USA
 */
#include <ff.h>
#if defined (_WIN32) && defined (FF_CPRINTF) // [
#include <stdarg.h>

int ff_cputwc(const wchar_t wc, FILE *f)
{
  HANDLE hConsole;

  if (stdout==f)
    hConsole=GetStdHandle(STD_OUTPUT_HANDLE);
  else if (stderr==f)
    hConsole=GetStdHandle(STD_ERROR_HANDLE);
  else
    return fputwc(wc,f);

  WriteConsoleW(
    hConsole,   // _In_             HANDLE  hConsoleOutput,
    &wc,        // _In_       const VOID    *lpBuffer,
    1,          // _In_             DWORD   nNumberOfCharsToWrite,
    NULL,       // _Out_            LPDWORD lpNumberOfCharsWritten,
    NULL        // _Reserved_       LPVOID  lpReserved
  );

  return wc;
}

int ff_cputws(const wchar_t *wcs, FILE *f)
{
  HANDLE hConsole;
  size_t len;
  DWORD lNumberOfCharsWritten;
  BOOL bWriteConsole;

  if (stdout==f)
    hConsole=GetStdHandle(STD_OUTPUT_HANDLE);
  else if (stderr==f)
    hConsole=GetStdHandle(STD_ERROR_HANDLE);
  else
    return fputws(wcs,f);

  len=wcslen(wcs);

  bWriteConsole=WriteConsoleW(
    hConsole,   // _In_             HANDLE  hConsoleOutput,
    wcs,        // _In_       const VOID    *lpBuffer,
    len,        // _In_             DWORD   nNumberOfCharsToWrite,
    &lNumberOfCharsWritten,
                // _Out_            LPDWORD lpNumberOfCharsWritten,
    NULL        // _Reserved_       LPVOID  lpReserved
  );

  return bWriteConsole?lNumberOfCharsWritten:0;
}

int ff_cvprintf(FILE *f, const wchar_t *format, va_list ap)
{
  HANDLE hConsole;
  size_t len;
  wchar_t *buf;

  if (stdout==f)
    hConsole=GetStdHandle(STD_OUTPUT_HANDLE);
  else if (stderr==f)
    hConsole=GetStdHandle(STD_ERROR_HANDLE);
  else
    return _vfwprintf_p(f,format,ap);

  len=_vswprintf_p(NULL,0,format,ap);
  buf=malloc((len+1)*sizeof buf[0]);

  if (!buf)
    return -1;

  _vswprintf_p(buf,len+1,format,ap);

  WriteConsoleW(
    hConsole,   // _In_             HANDLE  hConsoleOutput,
    buf,        // _In_       const VOID    *lpBuffer,
    len,        // _In_             DWORD   nNumberOfCharsToWrite,
    NULL,       // _Out_            LPDWORD lpNumberOfCharsWritten,
    NULL        // _Reserved_       LPVOID  lpReserved
  );

  free(buf);

  return len;
}

int ff_cprintf(FILE *f, const wchar_t *format, ...)
{
  va_list ap;
  int err;

  va_start(ap,format);
  err=ff_cvprintf(f,format,ap);
  va_end(ap);

  return err;
}
#endif // ]
